import os
import subprocess
from pathlib import Path
from shutil import copyfile

from damsel_diary.entry import Entry
from damsel_diary.exceptions import (
    InvalidEntry,
    InvalidHeadingLevel,
    UnknownTemplate,
)
from damsel_diary.parsing import string_split_newlines
from damsel_diary.time_format import TimeFormat


class DiaryWriter:
    """Writes entries to files."""

    def __init__(self, parent_path: Path = Path.home()) -> None:
        """Initialize an instance of DiaryWriter.

        Args:
            parent_path (Path, optional): The parent path of
            .damsel-diary/. Defaults to Path.home() (the home
            directory).
        """
        self.parent_path = parent_path
        self.time = TimeFormat()

        self.__build_tree()  # Build the expected directory tree.

    def __build_tree(self) -> None:
        """Build the expected directory tree."""
        # Create the diary directory.
        self.diary_path = self.parent_path.joinpath(".damsel-diary/")
        self.diary_path.mkdir(parents=True, exist_ok=True)

        self.templates_path = self.diary_path.joinpath("templates/")
        self.templates_path.mkdir(parents=True, exist_ok=True)

        templates = Path(__file__).parent.resolve().joinpath("templates/")

        for child in templates.iterdir():
            user_child = self.templates_path.joinpath(child.name)

            if user_child.exists() is False:
                copyfile(child, self.templates_path.joinpath(child.name))

    def today(
        self,
        amend: bool = False,
        template_name: str = "default",
        subject: str = "",
        new_heading_level: int = 2,
    ) -> Path:
        """Edit today's diary entry.

        Args:
            amend (bool, optional): Amend the entry without including
            the time heading. Defaults to False.
            template_name (str, optional): The name of the template to
            use. Defaults to "default".
            new_heading_level (int): The level of the new Markdown
            heading. Defaults to 2.

        Returns:
            Path: The path to the new entry.
        """
        # Create the expected path to today's entry.
        today_path = self.diary_path.joinpath(
            f"{self.time.year}/{self.time.month_num}/"
        )
        today_path.mkdir(parents=True, exist_ok=True)

        # Check for valid heading level.
        new_heading_level = self.__check_heading_level(new_heading_level)

        # Find template.
        template_path = self.__find_template(template_name)

        # Create today's entry.
        entry_path = today_path.joinpath(f"{self.time.day_num}.md")
        entry = Entry(entry_path, template_path, self.time)

        # Evaluate the entry file - if it exists and is not empty, then
        # extend it.
        if entry.file_path.is_file() and entry.file_path.stat().st_size != 0:
            self.__extend_entry(entry, amend, subject, new_heading_level)

        # Else, create a new entry.
        else:
            self.__new_entry(entry, amend, subject, new_heading_level)

        return entry.file_path

    def __find_template(self, template_name: str) -> Path:
        """Find the template in .damsel-diary/templates/.

        Args:
            template_name (str): The name of the template.

        Raises:
            UnknownTemplate: Raised if the template cannot be found.

        Returns:
            Path: The path to the template.
        """
        # Create a path of the name.
        template_path = self.templates_path.joinpath(f"{template_name}.md")

        if template_path.is_file():
            return template_path

        else:
            raise UnknownTemplate(f"{template_path}: cannot find template")

    @classmethod
    def __extend_entry(
        cls, entry: "Entry", amend: bool, subject: str, new_heading_level: int
    ) -> None:
        """Extend an existing entry.

        Args:
            entry (Entry): An entry.
            amend (bool): Flag to amend the entry without including a
            time heading. If True, a time heading is not added.
            new_heading_level (int): The level of the new Markdown
            heading.
        """
        try:
            with open(entry.file_path, "a+") as f:

                current_lines = string_split_newlines(entry.current_contents)

                # The result should have an empty line between the
                # original contents and the extension
                try:
                    if (current_lines[len(current_lines) - 1] == "\n") and (
                        current_lines[len(current_lines) - 2] != "\n"
                    ):
                        f.write("\n")

                except IndexError:
                    pass

                if amend is False:
                    # Write the time heading.
                    f.write(
                        cls.__new_heading(entry, subject, new_heading_level)
                    )

                # If a template has been chosen, add it to the end of
                # the file.
                if entry.parsed_template != "":

                    f.write(entry.parsed_template)

        except OSError:
            raise InvalidEntry(f"{entry.file_path.name}: cannot be accessed")

    @classmethod
    def __new_entry(
        cls, entry: "Entry", amend: bool, subject: str, new_heading_level: int
    ) -> None:
        """Create a new entry.

        Args:
            entry (Entry): An entry.
            amend (bool): Flag to amend the entry without including a
            time heading. If True, a time heading is not added.
            new_heading_level (int): The level of the new Markdown
            heading.
        """
        try:
            with open(entry.file_path, "w") as f:

                # Write the title (date).
                f.write(
                    "# "
                    + entry.time_format.date
                    + ": "
                    + entry.time_format.day_name
                    + (2 * "\n")
                )

                if amend is False:
                    # Write the time heading.
                    f.write(
                        cls.__new_heading(entry, subject, new_heading_level)
                    )

                # If a template has been chosen, add it to the end of
                # the file.
                if entry.parsed_template != "":
                    f.write(entry.parsed_template)

        except OSError:
            raise InvalidEntry(f"{entry.file_path.name}: cannot be accessed")

    @staticmethod
    def __new_heading(
        entry: "Entry", subject: str, new_heading_level: int
    ) -> str:
        """Create a new time heading.

        Args:
            entry (Entry): An entry.
            new_heading_level (int): The level of the new Markdown
            heading.

        Returns:
            str: The time heading in Markdown format.
        """
        return (
            "#" * new_heading_level
            + " "
            + entry.time_format.time
            + " "
            + subject
            + (2 * "\n")
        )

    @staticmethod
    def open_entry(path: Path) -> None:
        """Open an entry using the $EDITOR variable.

        Args:
            path (Path): The path to the entry.
        """
        subprocess.call([os.environ["EDITOR"], str(path)])

    @staticmethod
    def __check_heading_level(new_heading_level: int) -> int:
        """Check that the heading level is within the expected interval.

        Args:
            new_heading_level (int): The Markdown heading level.

        Raises:
            argparse.ArgumentTypeError: Raised if hl is not in the
            interval [2, 6].

        Returns:
            int: The new heading level.
        """
        # Check that l is in the valid range for Markdown headings
        # (excluding the title).
        if 2 <= new_heading_level <= 6:
            return new_heading_level

        else:
            raise InvalidHeadingLevel(
                f"{new_heading_level}: heading-level must be in the interval "
                "[2, 6]"
            )
